# -*- coding: utf-8 -*-
# --------------------------------------------------------------------
# Copyright © 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

import os
import tempfile
import tarfile
import shutil
import unittest

# file mode to use for creating test directories.
TEST_DIR_MODE = 0o750

script_name = os.path.basename(__file__)


def make_tmp_dir(tag=None):
    '''
    Create a temporary directory.
    '''

    if tag:
        prefix = '{}-{}-'.format(script_name, tag)
    else:
        prefix = script_name

    return tempfile.mkdtemp(prefix=prefix)


def append_file(path, contents):
    '''
    Append to a regular file (create it doesn't exist).
    '''

    dirname = os.path.dirname(path)
    os.makedirs(dirname, mode=TEST_DIR_MODE, exist_ok=True)

    with open(path, 'a') as fh:
        if contents:
            fh.writelines(contents)

            if not contents.endswith('\n'):
                fh.write('\n')


def create_file(path, contents):
    '''
    Create a regular file.
    '''
    append_file(path, contents)


def mock_get_root_partitions_by_label():
    '''
    Fake disk partition details for testing.
    '''
    matches = []

    matches.append(('system-a', '/dev/sda3', '/'))
    matches.append(('system-b', '/dev/sda4', '/writable/cache/system'))

    return matches


class UpdateTree():
    '''
    Representation of a directory tree that will be converted into an
    update archive.
    '''
    TEST_REMOVED_FILE = 'removed'
    TEST_SYSTEM_DIR = 'system/'

    def __init__(self):

        # Directory tree used to construct the tar file from.
        # Also used to hold the TEST_REMOVED_FILE file.
        self.dir = make_tmp_dir(tag='UpdateTree-tar-source')

        self.removed_file = os.path.join(self.dir, self.TEST_REMOVED_FILE)

        # Directory to place create/modify files into.
        self.system_dir = os.path.join(self.dir, self.TEST_SYSTEM_DIR)

        # Directory used to write the generated tarfile to.
        # This directory should also be used to write the command file
        # to.
        self.tmp_dir = make_tmp_dir(tag='UpdateTree-cache')

    def destroy(self):
        if os.path.exists(self.dir):
            shutil.rmtree(self.dir)

        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def add_to_removed_file(self, removed_files):
        '''
        Add the specified list of files to the removed file.

        The 'removed' file is simply a file with a well-known name that
        contains a list of files (one per line) to be removed from a
        system before the rest of the update archive is unpacked.

        :param removed_files: list of file names to add to the removed file.

        '''
        # all files listed in the removed list must be system files
        final = list(map(lambda a:
                     os.path.normpath('{}{}'.format(self.TEST_SYSTEM_DIR,
                                                    a)), removed_files))

        contents = "".join(final)
        append_file(self.removed_file, contents)

    def tar_filter(self, member):
        '''
        Function to filter the tarinfo members before creating the
        archive.
        '''
        # members are created with relative paths (no leading slash)
        path = os.sep + member.name

        if member.name == '/.':
            return None

        i = path.find(self.dir)
        assert(i == 0)

        # remove the temporary directory elements
        # (+1 for the os.sep we added above)
        member.name = path[len(self.dir)+1:]

        return member

    def create_archive(self, name):
        '''
        Create an archive with the specified name from the UpdateTree
        object. Also creates a fake signature file alongside the archive
        file since this is currently required by the upgrader (although
        it is not validated).

        :param name: name of tarfile.
        :param name: full path to xz archive to create.
        :return full path to tar file with name @name.
        '''

        self.tar_path = os.path.join(self.tmp_dir, name)
        tar = tarfile.open(self.tar_path, 'w:xz')

        # We can't just add recursively since that would attempt to add
        # the parent directory. However, the real update tars don't
        # include that, and attempting to ignore the parent directory
        # results in an empty archive. So, walk the tree and add
        # file-by-file.
        for path, names, files in os.walk(self.dir):
            for file in files:
                full = os.path.join(path, file)
                tar.add(full, recursive=False, filter=self.tar_filter)
            if not files and not names:
                # add (empty) directories
                tar.add(path, recursive=False, filter=self.tar_filter)

        tar.close()

        signature = '{}.asc'.format(self.tar_path)

        with open(signature, 'w') as fh:
            fh.write('fake signature file')

        return self.tar_path


class UbuntuCoreUpgraderTestCase(unittest.TestCase):
    '''
    Base class for Upgrader tests.

    Most of the tests follow a standard pattern:

    1) Create an UpdateTree object:

         update = UpdateTree()

       This creates 2 temporary directories:

       - self.system_dir: Used to generate an update archive from.

       - self.tmp_dir: Used to write the generated archive file to. The
         intention is that this directory should also be used to hold
         the command file.

    2) Removal tests call update.add_to_removed_file(file) to add a
       particular file to the removals file in the update archive.

    3) Create/Modify tests create files below update.system_dir.

    4) Create an update archive (which includes the removals file
       and all files below update.system_dir):

         archive = update.create_archive(self.TARFILE)

    5) Create a command file (which tells the upgrader what to do
       and which archive files to apply):

         make_command_file(...)

    6) Create a victim directory. This is a temporary directory where
       the upgrade will happen.

    7) Start the upgrade:

         call_upgrader(...)

    8) Perform checks on the victim directory to ensure that upgrade
       did what was expected.

    '''

    TARFILE = 'update.tar.xz'

    # result of last test run. Hack to deal with fact that even if a
    # test fails, unittest still calls .tearDown() (whomever thought
    # that was a good idea...?)
    currentResult = None

    def setUp(self):
        '''
        Test setup.
        '''
        # Create an object to hold the tree that will be converted into
        # an upgrade archive.
        self.update = UpdateTree()

        # The directory which will have the update archive applied to
        # it.
        self.victim_dir_base = make_tmp_dir(tag='victim')
        self.victim_dir = os.path.join(self.victim_dir_base, 'system')
        os.makedirs(self.victim_dir, mode=0o750)

    def tearDown(self):
        '''
        Test cleanup.
        '''

        if not self.currentResult.wasSuccessful():
            # Do not clean up - the only sane option if a test fails.
            return

        self.update.destroy()
        self.update = None

        shutil.rmtree(self.victim_dir)
        self.victim_dir = None

    def run(self, result=None):
        self.currentResult = result
        unittest.TestCase.run(self, result)



import json
import os
import select
import subprocess
import sys

from systemimage.config import config
config  # pyflakes


class ApplyUpgradeError(Exception):
    pass


def is_sym_link_broken(path):
    '''
    Returns True if the specified path is a broken symbolic link, else
    False.
    '''
    try:
        os.lstat(path)
        os.stat(path)
    except:
        return True
    return False


class ApplyUpgrade:
    """Hook for system-image-cli"""

    UPGRADER = ["ubuntu-core-upgrade"]
    UPGRADER_ARGS = ["/writable/cache/ubuntu_command"]

    def apply_update(self):
        p = subprocess.Popen(
            self.UPGRADER + self.UPGRADER_ARGS,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        # collect the output while iterating to avoid hitting the 64k
        # pipe size limit (unlikely but...)
        stdout = ""
        while p.poll() is None:
            # send alive ping
            sys.stdout.write(json.dumps({"type": "spinner",
                                         "msg": "Applying update"})+"\n")
            sys.stdout.flush()
            if select.select([p.stdout.fileno()], [], [], 0.1) != ([], [], []):
                stdout += p.stdout.read(1).decode("utf-8", errors="replace")
        # get return code
        ret_code = p.poll()
        # close the pipes
        p.communicate()
        if ret_code != 0:
            raise ApplyUpgradeError(
                "%s exited with %s:\n%s", self.UPGRADER, ret_code, stdout)
        return True

    # entry point for s-i 2.x
    def reboot(self):
        self.apply_update()

    # entry point for s-i 3.0
    def apply(self):
        self.apply_update()



import ctypes
import os


class UnshareFlags:
    # from /usr/include/linux/sched.h, see also unshare(2)
    CLONE_FILES = 0x00000400
    CLONE_FS = 0x00000200
    CLONE_NEWIPC = 0x08000000
    CLONE_NEWNET = 0x40000000
    CLONE_NEWNS = 0x00020000  # equivalent to unshare --mount
    CLONE_NEWUTS = 0x04000000
    CLONE_SYSVSEM = 0x00040000
    CLONE_NEWPID = 0x20000000      # /* New pid namespace */


def unshare(flags):
    libc = ctypes.CDLL("libc.so.6", use_errno=True)
    res = libc.unshare(ctypes.c_int(flags))
    if res == -1:
        e = ctypes.get_errno()
        raise OSError(e, os.strerror(e))



#!/usr/bin/python

import unittest
from unittest.mock import (
    patch,
)

from ubuntucoreupgrader.apply import (
    ApplyUpgrade,
    ApplyUpgradeError,
)


class ApplyUpgradeTestCase(unittest.TestCase):

    def test_apply_update_success(self):
        with patch("ubuntucoreupgrader.apply.config"):
            a = ApplyUpgrade()
            a.UPGRADER = ["true"]
            res = a.apply_update()
            self.assertEqual(res, True)

    def test_apply_update_fail(self):
        with patch("ubuntucoreupgrader.apply.config"):
            a = ApplyUpgrade()
            a.UPGRADER = ["false"]
            with self.assertRaises(ApplyUpgradeError):
                res = a.apply_update()
                self.assertEqual(res, False)

    def test_apply_update_generates_loads_of_data(self):
        with patch("ubuntucoreupgrader.apply.config"):
            a = ApplyUpgrade()
            a.UPGRADER = ["dd", "if=/dev/urandom", "bs=1k", "count=65"]
            a.UPGRADER_ARGS = []
            res = a.apply_update()
            self.assertEqual(res, True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# --------------------------------------------------------------------
# Copyright © 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

# --------------------------------------------------------------------
# Terminology:
#
# - "update archive": a fake system-image "ubuntu-hash.tar.xz" tar
#   archive.
# --------------------------------------------------------------------

import tarfile
import tempfile
import unittest
from unittest.mock import patch
import os
import shutil
import sys

from ubuntucoreupgrader.upgrader import (
    tar_generator,
    Upgrader,
    parse_args,
)

base_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.append(base_dir)

from ubuntucoreupgrader.tests.utils import (
    create_file,
    make_tmp_dir,
    UbuntuCoreUpgraderTestCase,
    )

# file mode to use for creating test directories.
TEST_DIR_MODE = 0o750

CMD_FILE = 'ubuntu_command'


def make_default_options():
    return parse_args([])


class UpgradeTestCase(unittest.TestCase):

    def test_tar_generator_unpack_assets(self):
        tempf = tempfile.TemporaryFile()
        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:

            # special top-level file that should not be unpacked
            tar.add(__file__, "removed")

            tar.add(__file__, "system/bin/true")
            tar.add(__file__, "assets/vmlinuz")
            tar.add(__file__, "assets/initrd.img")
            tar.add(__file__, "unreleated/something")
            tar.add(__file__, "hardware.yaml")

            result = [m.name for m in tar_generator(tar, "cache_dir", [], "/")]
            self.assertEqual(result, ["system/bin/true", "assets/vmlinuz",
                                      "assets/initrd.img", "hardware.yaml"])

    def test_tar_generator_system_files_unpack(self):
        tempf = tempfile.TemporaryFile()
        root_dir = make_tmp_dir()
        cache_dir = make_tmp_dir()

        sys_dir = os.path.join(cache_dir, 'system')

        os.makedirs(os.path.join(root_dir, 'dev'))
        os.makedirs(sys_dir)

        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:
            tar.add(__file__, "assets/vmlinuz")
            tar.add(__file__, "assets/initrd.img")

            device_a = '/dev/null'
            path = (os.path.normpath('{}/{}'.format(sys_dir, device_a)))
            touch_file(path)

            self.assertTrue(os.path.exists(os.path.join(root_dir, device_a)))

            # should not be unpacked as already exists
            tar.add(__file__, "system{}".format(device_a))

            device_b = '/dev/does-not-exist'
            self.assertFalse(os.path.exists(os.path.join(root_dir, device_b)))

            # should be unpacked as does not exist
            tar.add(__file__, "system{}".format(device_b))

            expected = ["assets/vmlinuz", "assets/initrd.img",
                        "dev/does-not-exist"]

            expected_results = [os.path.join(root_dir, file)
                                for file in expected]

            result = [m.name for m in
                      tar_generator(tar, cache_dir, [], root_dir)]

            self.assertEqual(result, expected_results)

        shutil.rmtree(root_dir)
        shutil.rmtree(cache_dir)

    def test_tar_generator_system_files_remove_before_unpack(self):
        """
        Test that the upgrader removes certain files just prior to
        overwriting them via the unpack.
        """

        tempf = tempfile.TemporaryFile()
        root_dir = make_tmp_dir()
        cache_dir = make_tmp_dir()
        sys_dir = os.path.join(cache_dir, 'system')

        os.makedirs(sys_dir)

        file = 'a/file'
        dir = 'a/dir'

        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:

            file_path = os.path.normpath('{}/{}'.format(root_dir, file))
            touch_file(file_path)
            self.assertTrue(os.path.exists(file_path))

            dir_path = os.path.normpath('{}/{}'.format(root_dir, dir))
            os.makedirs(dir_path)
            self.assertTrue(os.path.exists(dir_path))

            tar.add(__file__, "system/{}".format(file))

            expected = [file]
            expected_results = [os.path.join(root_dir, f)
                                for f in expected]

            result = [m.name for m in
                      tar_generator(tar, cache_dir, [], root_dir)]

            self.assertEqual(result, expected_results)

            # file should be removed
            self.assertFalse(os.path.exists(file_path))

            # directory should not be removed
            self.assertTrue(os.path.exists(dir_path))

        shutil.rmtree(root_dir)
        shutil.rmtree(cache_dir)


def touch_file(path):
    '''
    Create an empty file (creating any necessary intermediate
    directories in @path).
    '''
    create_file(path, None)


def make_commands(update_list):
    """
    Convert the specified list of update archives into a list of command
    file commands.
    """
    l = []

    # we don't currently add a mount verb (which would be more
    # realistic) since we can't handle that in the tests.
    # ##l.append('mount system')

    for file in update_list:
        l.append('update {} {}.asc'.format(file, file))

    # we don't currently add an unmount verb (which would be more
    # realistic) since we can't handle that in the tests.
    # ##l.append('unmount system')

    return l


class UbuntuCoreUpgraderObectTestCase(UbuntuCoreUpgraderTestCase):

    def test_create_object(self):
        root_dir = make_tmp_dir()

        file = 'created-regular-file'

        file_path = os.path.join(self.update.system_dir, file)
        create_file(file_path, 'foo bar')

        self.cache_dir = self.update.tmp_dir

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        dest = os.path.join(self.cache_dir, os.path.basename(archive))
        touch_file('{}.asc'.format(dest))

        commands = make_commands([self.TARFILE])

        options = make_default_options()

        # XXX: doesn't actually exist, but the option must be set since
        # the upgrader looks for the update archives in the directory
        # where this file is claimed to be.
        options.cmdfile = os.path.join(self.cache_dir, 'ubuntu_command')

        options.root_dir = root_dir

        upgrader = Upgrader(options, commands, [])
        upgrader.MOUNTPOINT_CMD = "true"
        upgrader.run()

        path = os.path.join(root_dir, file)
        self.assertTrue(os.path.exists(path))

        shutil.rmtree(root_dir)

    @patch('ubuntucoreupgrader.upgrader.get_mount_details')
    @patch('ubuntucoreupgrader.upgrader.mount')
    @patch('ubuntucoreupgrader.upgrader.unmount')
    def test_no_format_in_cmd(self, mock_umount, mock_mount,
                              mock_mount_details):

        # If the command file does not contain the format command, mkfs
        # should not be called.
        with patch('ubuntucoreupgrader.upgrader.mkfs') as mock_mkfs:
            args = ['cmdfile']
            options = parse_args(args=args)
            commands = make_commands([self.TARFILE])
            upgrader = Upgrader(options, commands, [])
            upgrader.TIMESTAMP_FILE = '/dev/null'
            upgrader.MOUNTPOINT_CMD = "true"
            upgrader.run()

        # No format command in command file, so should not have been
        # called.
        self.assertFalse(mock_mkfs.called)

    @patch('ubuntucoreupgrader.upgrader.get_mount_details')
    @patch('ubuntucoreupgrader.upgrader.mount')
    @patch('ubuntucoreupgrader.upgrader.unmount')
    def test_format(self, mock_umount, mock_mount, mock_mount_details):
        MOCK_FS_TUPLE = ("device", "fstype", "label")
        mock_mount_details.return_value = MOCK_FS_TUPLE

        # mkfs should be called if the format command is specified in
        # the command file.
        with patch('ubuntucoreupgrader.upgrader.mkfs') as mock_mkfs:
            args = ['cmdfile']
            options = parse_args(args=args)
            commands = make_commands([self.TARFILE])

            # add format command to command file
            commands.insert(0, 'format system')

            upgrader = Upgrader(options, commands, [])
            upgrader.TIMESTAMP_FILE = '/dev/null'
            upgrader.MOUNTPOINT_CMD = "true"
            upgrader.run()
            self.assertTrue(upgrader.other_has_been_formatted)

        mock_mkfs.assert_called_with(*MOCK_FS_TUPLE)

    @patch.object(Upgrader, 'sync_partitions')
    @patch('ubuntucoreupgrader.upgrader.get_mount_details')
    @patch('ubuntucoreupgrader.upgrader.remount')
    @patch('ubuntucoreupgrader.upgrader.mount')
    @patch('ubuntucoreupgrader.upgrader.fsck')
    @patch('ubuntucoreupgrader.upgrader.mkfs')
    @patch('ubuntucoreupgrader.upgrader.unmount')
    def test_mount_unmount_no_format(self, mock_umount, mock_mkfs, mock_fsck,
                                     mock_mount, mock_remount,
                                     mock_mount_details, mock_sync_partitions):
        MOCK_FS_TUPLE = ("device", "fstype", "label")
        mock_mount_details.return_value = MOCK_FS_TUPLE

        # mkfs should be called if the format command is specified in
        # the command file.
        args = ['cmdfile']
        options = parse_args(args=args)
        commands = make_commands([self.TARFILE])

        commands.insert(0, 'mount system')
        commands.append('unmount system')

        upgrader = Upgrader(options, commands, [])
        upgrader.TIMESTAMP_FILE = '/dev/null'
        upgrader.MOUNTPOINT_CMD = "true"
        upgrader.run()

        self.assertTrue(mock_sync_partitions.called)

    @patch.object(Upgrader, 'sync_partitions')
    @patch('ubuntucoreupgrader.upgrader.get_mount_details')
    @patch('ubuntucoreupgrader.upgrader.remount')
    @patch('ubuntucoreupgrader.upgrader.mount')
    @patch('ubuntucoreupgrader.upgrader.fsck')
    @patch('ubuntucoreupgrader.upgrader.mkfs')
    @patch('ubuntucoreupgrader.upgrader.unmount')
    def test_mount_unmount_with_format(self, mock_umount, mock_mkfs, mock_fsck,
                                       mock_mount, mock_remount,
                                       mock_mount_details,
                                       mock_sync_partitions):
        MOCK_FS_TUPLE = ("device", "fstype", "label")
        mock_mount_details.return_value = MOCK_FS_TUPLE

        # mkfs should be called if the format command is specified in
        # the command file.
        args = ['cmdfile']
        options = parse_args(args=args)
        commands = make_commands([self.TARFILE])

        commands.insert(0, 'format system')
        commands.insert(1, 'mount system')
        commands.append('unmount system')

        upgrader = Upgrader(options, commands, [])
        upgrader.TIMESTAMP_FILE = '/dev/null'
        upgrader.MOUNTPOINT_CMD = "true"
        upgrader.run()

        self.assertFalse(mock_sync_partitions.called)

    def test_empty_removed_file(self):
        root_dir = make_tmp_dir()

        file = 'created-regular-file'

        file_path = os.path.join(self.update.system_dir, file)
        create_file(file_path, 'foo bar')

        self.cache_dir = self.update.tmp_dir

        removed_file = self.update.removed_file
        # Create an empty removed file
        create_file(removed_file, '')

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        dest = os.path.join(self.cache_dir, os.path.basename(archive))
        touch_file('{}.asc'.format(dest))

        commands = make_commands([self.TARFILE])

        options = make_default_options()

        # XXX: doesn't actually exist, but the option must be set since
        # the upgrader looks for the update archives in the directory
        # where this file is claimed to be.
        options.cmdfile = os.path.join(self.cache_dir, 'ubuntu_command')

        options.root_dir = root_dir

        upgrader = Upgrader(options, commands, [])
        upgrader.cache_dir = self.cache_dir
        upgrader.MOUNTPOINT_CMD = "true"

        si_file = self.TARFILE
        si_signature = "{}.asc".format(self.TARFILE)
        upgrader._cmd_update([si_file, si_signature])

        # Ensure that the upgrader has not attempted to remove the
        # cache_dir when the 'removed' file is empty.
        #
        # (Regression test for LP: #1437225).
        self.assertTrue(os.path.exists(upgrader.cache_dir))

        shutil.rmtree(self.cache_dir)

    def test_other_considered_empty(self):

        si_config_file = '/etc/system-image/channel.ini'

        cache_dir = make_tmp_dir()
        args = ['cmdfile']
        options = parse_args(args=args)
        commands = make_commands([self.TARFILE])

        upgrader = Upgrader(options, commands, [])
        upgrader.TIMESTAMP_FILE = '/dev/null'
        upgrader.MOUNTPOINT_CMD = "true"
        upgrader.cache_dir = cache_dir

        self.assertTrue(upgrader.other_considered_empty())

        target = upgrader.get_mount_target()
        channel_ini = os.path.normpath('{}/{}'.format(target,
                                       si_config_file))
        si_dir = os.path.dirname(channel_ini)

        os.makedirs(si_dir)

        self.assertTrue(upgrader.other_considered_empty())

        touch_file(channel_ini)

        self.assertTrue(upgrader.other_considered_empty())

        create_file(channel_ini, "my size is >0")

        self.assertFalse(upgrader.other_considered_empty())

        os.remove(channel_ini)

        self.assertTrue(upgrader.other_considered_empty())

        shutil.rmtree(cache_dir)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3

from setuptools import setup

setup(name='ubuntucoreupgrader',
      packages=['ubuntucoreupgrader'],
      test_suite="ubuntucoreupgrader.tests",
      scripts=['bin/ubuntu-core-upgrade'],
      setup_requires=[
          "flake8",
      ],
      )



# --------------------------------------------------------------------
# Copyright © 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

# --------------------------------------------------------------------
# TODO:
#
# - Stop services, unpack, restart services?
# -   XXX: *DO* unpack device files once above handled.
# - Fix bug triggered by restarting systemd which causes
#   Upgrader.systemd.connection* to hang for 30 seconds even when
#   recreated.
# --------------------------------------------------------------------

'''
Apply an Ubuntu Core system image update.

WARNING : If running an in-place upgrade, should be run within a new
          mount namespace to avoid disturbing the existing (read-only)
          root filesystem.

NOTE    : The signature files associated with each image file is *NOT*
          re-verified, since this must already have been handled by
          system-image-cli(1).
'''

import sys
import os
import logging
import shutil
import subprocess
import tempfile
import tarfile
import argparse

script_name = os.path.basename(__file__)

log = logging.getLogger()

DEFAULT_ROOT = '/'

# The tar file contains a 'system/' directory with the new files
# to apply to the real system. It may also contain a top-level
# file called 'removed' that lists files relative to the system
# mount for files on the system that should be removed before
# unpacking the rest of the archive.
#
# Examples:
#
# - if the file '/etc/foo' should be removed, it
#   would be specified in the removed_file as 'system/etc/foo'.
#
# - if the file '/etc/bar' should be modified, it
#   would be specified in the tar file as member 'system/etc/bar'.
TAR_FILE_SYSTEM_PREFIX = 'system/'

# boot assets (kernel, initrd, .dtb's)
TAR_FILE_ASSETS_PREFIX = 'assets/'

# the file that describes the boot assets
TAR_FILE_HARDWARE_YAML = 'hardware.yaml'

TAR_FILE_REMOVED_FILE = 'removed'

SYSTEM_IMAGE_CHANNEL_CONFIG = '/etc/system-image/channel.ini'


def parse_args(args):
    '''
    Handle command-line options.

    Returns: dict of command-line options.
    '''

    parser = argparse.ArgumentParser(description='System image Upgrader')

    parser.add_argument(
        '--debug',
        nargs='?', const=1, default=0, type=int,
        help='Dump debug info (specify numeric value to increase verbosity)')

    parser.add_argument(
        '-n', '--dry-run',
        action='store_true',
        help='''
        Simulate an update including showing processes that have locks
        on files
        ''')

    parser.add_argument(
        '--leave-files',
        action='store_true',
        help='Do not remove the downloaded system image files after upgrade')

    parser.add_argument(
        '--root-dir',
        default=DEFAULT_ROOT,
        help='Specify an alternative root directory (for testing ONLY)')

    parser.add_argument(
        '-t', '--tmpdir',
        help='Specify name for pre-existing temporary directory to use')

    parser.add_argument(
        'cmdfile', action="store",
        nargs='?',
        help='Name of file containing commands to execute')

    return parser.parse_args(args=args)


def remove_prefix(path, prefix=TAR_FILE_SYSTEM_PREFIX):
    '''
    Remove specified prefix from path and return the result.

    Prefix must end with a slash.

    If @prefix is not a prefix of @path, returns None.
    '''

    assert(prefix.endswith('/'))

    i = path.find(prefix)

    if i < 0:
        return None

    # ensure that the returned path has a leading slash
    return path[len(prefix)-1:]


def fsck(device):
    '''
    Run fsck(8) on specified device.
    '''
    assert (os.path.exists(device))

    failed = False

    cmd = '/sbin/fsck'
    args = []

    args.append(cmd)

    # Paranoia - don't fsck if already mounted
    args.append('-M')

    args.append('-av')
    args.append(device)

    log.debug('running: {}'.format(args))

    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)

    ret = proc.wait()

    if ret == 0:
        return

    stdout, stderr = proc.communicate()

    # fsck's return code 1 means: "Filesystem errors corrected"
    # (aka a warning - FS is consistent [now]).
    failed = False if ret == 1 else True

    log.error('{} returned {} ({}): {}, {}'
              .format(args,
                      ret,
                      "failed" if failed else "warning",
                      stdout,
                      stderr))

    if failed:
        sys.exit(1)


def mkfs(device, fs_type, label):
    '''
    Run mkfs(8) on specified device.
    '''
    assert (os.path.exists(device))

    args = ['/sbin/mkfs',
            '-v',
            '-t', fs_type,
            '-L', label,
            device]

    log.debug('running: {}'.format(args))

    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)

    ret = proc.wait()

    if ret == 0:
        return

    stdout, stderr = proc.communicate()

    log.error('{} returned {}: {}, {}'
              .format(args,
                      ret,
                      stdout,
                      stderr))
    sys.exit(1)


def touch_file(path):
    '''
    Create an empty file specified by path.
    '''
    open(path, 'w').close()


def get_mount_details(target):
    '''
    Returns a list comprising device, filesystem type and filesystem
    label for the mount target specified.
    '''
    args = ['findmnt', '-o', 'SOURCE,FSTYPE,LABEL', '-n', target]
    output = subprocess.check_output(args, universal_newlines=True)
    output = output.strip().split()

    if len(output) != 3:
        sys.exit('Failed to determine mount details for {}'
                 .format(target))

    return output


def remount(mountpoint, options):
    """
    Remount mountpoint using the specified options string (which is
    passed direct to the mount command and which must not contain
    spaces).
    """
    cmd = 'mount'

    args = []

    args.append(cmd)
    args.append('-oremount,{}'.format(options))
    args.append(mountpoint)

    log.debug('running: {}'.format(args))

    proc = subprocess.Popen(args,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    if proc.wait() != 0:
        stderr = proc.communicate()[1]
        log.error('failed to remount ({}): {}'
                  .format(args, stderr))
        sys.exit(1)


def mount(source, target, options=None):
    '''
    Mount @source on @target using optional specified
    options string (which must not contain spaces).
    '''

    cmd = 'mount'

    args = []

    args.append(cmd)
    if options:
        args += ['-o', options]
    args.append(source)
    args.append(target)

    log.debug('running: {}'.format(args))

    proc = subprocess.Popen(args,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    if proc.wait() != 0:
        stderr = proc.communicate()[1]
        log.error('failed to mount ({}): {}'
                  .format(args, stderr))
        sys.exit(1)


def bind_mount(source, target):
    '''
    Bind mount @source to @target.
    '''
    log.debug('bind mounting existing root')
    mount(source, target, 'bind')


def unmount(target, options=None):
    '''
    Unmount the specified mount target, using the specified list of
    options.
    '''

    log.debug('unmounting {}'.format(target))

    args = []
    args.append('umount')
    if options:
        args.extend(options)
    args.append(target)

    log.debug('running: {}'.format(args))

    proc = subprocess.Popen(args,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)

    if proc.wait() != 0:
        stderr = proc.communicate()[1]
        log.error('failed to unmount ({}): {}'
                  .format(args, stderr))

        # its a lazy umount, do not fail hard just now.
        #
        # FIXME: entry in mount table for /etc/writable/hostname really
        # does not seem to have been mounted by systemd. No problem when
        # booting under upstart as mountall does bind mount that path.
        #
        # sys.exit(1)


def tar_generator(tar, cache_dir, removed_files, root_dir):
    '''
    Generator function to handle extracting members from the system
    image tar files.
    '''
    for member in tar:

        # Restrictions:
        #
        # - Don't unpack the removed file.
        # - Don't unpack device files *iff* they already exist.
        # - Don't unpack files that are not below
        #   TAR_FILE_SYSTEM_PREFIX or TAR_FILE_ASSETS_PREFIX.
        #
        # - XXX: This is a temporary function to work-around for
        #   LP: #1381121: we shouldn't need to filter which files are
        #   extracted!
        device_prefix = '{}dev/'.format(TAR_FILE_SYSTEM_PREFIX)

        # The partition to update is mounted at "cache_dir/system".
        # However, the unpack occurs in directory cache_dir (since the
        # tar files prefixes all system files with
        # TAR_FILE_SYSTEM_PREFIX). For example, member.name may be
        # something like "system/dev/null".
        mount_path = '{}/{}'.format(cache_dir, member.name)

        unpack = True

        if member.name == removed_files:
            # already handled
            unpack = False

        if member.name.startswith(device_prefix) and \
                os.path.exists(mount_path):
            # device already exists
            unpack = False

        member_prefix = None

        if member.name.startswith(TAR_FILE_SYSTEM_PREFIX):
            member_prefix = TAR_FILE_SYSTEM_PREFIX
        elif member.name.startswith(TAR_FILE_ASSETS_PREFIX):
            member_prefix = TAR_FILE_ASSETS_PREFIX

        if not member_prefix:
            # unrecognised prefix directory
            unpack = False

        if member.name == TAR_FILE_HARDWARE_YAML:
            unpack = True

        if not unpack:
            log.debug('not unpacking file {}'.format(member.name))
            continue

        # A modified root directory requires convering
        # absolute paths to be located below the modified root
        # directory.
        if root_dir != '/':
            if member_prefix == TAR_FILE_SYSTEM_PREFIX:
                base = remove_prefix(member.name, prefix=member_prefix)
                member.name = os.path.normpath('{}/{}'.format(root_dir,
                                                              base))
            else:
                member.name = os.path.normpath('{}/{}'.format(root_dir,
                                                              member.name))

            if member.type in (tarfile.LNKTYPE, tarfile.SYMTYPE) \
                    and member.linkname.startswith('/'):
                # Hard and symbolic links also need their
                # 'source' updated to take account of the root
                # directory.
                #
                # But rather than remove the prefix, we add the
                # root directory as a prefix to contain the link
                # within that root.
                base = os.path.join(root_dir, member.linkname)

                member.linkname = '{}{}'.format(root_dir, base)

            path = member.name
        else:
            path = mount_path

        log.debug('unpacking file {} with path {}'
                  .format(member.name, path))

        # If the file is a binary and that binary is currently
        # being executed by a process, attempting to unpack it
        # will result in ETXTBSY (OSError: [Errno 26] Text file
        # busy). The simplest way around this issue is to unlink
        # the file just before unpacking it (ideally, we'd catch
        # the exception and handle it separately). This allows
        # the unpack to continue, and the running process to
        # continue to use it's (old) version of the binary until
        # it's corresponding service is restarted.
        #
        # Note that at this point, we already have another copy
        # of the inode below /lost+found/.
        if not member.isdir() and os.path.lexists(path):
            log.debug('removing file {}'.format(path))
            os.unlink(path)

        yield member


class Upgrader():

    # FIXME: Should query system-image-cli (see bug LP:#1380574).
    DEFAULT_CACHE_DIR = '/writable/cache'

    # Directory to mount writable root filesystem below the cache
    # diretory.
    MOUNT_TARGET = 'system'

    # Magic file that records time last system image was successfully applied.
    TIMESTAMP_FILE = '.last_update'

    DIR_MODE = 0o750

    MOUNTPOINT_CMD = "mountpoint"

    def __init__(self, options, commands, remove_list):
        """
        @options list of command-line options.
        @commands: array of declarative commands to run to apply the new
         s-i rootfs update.
        @remove_list: list of files to remove before applying the new
         image.

        """
        self.dispatcher = {
            'format': self._cmd_format,
            'load_keyring': self._cmd_load_keyring,
            'mount': self._cmd_mount,
            'unmount': self._cmd_unmount,
            'update': self._cmd_update,
        }

        self.options = options

        assert('root_dir' in self.options)

        # array of imperative commands to run, as generated by
        # system-image-cli(1).
        self.commands = commands

        # files to remove before unpacking new image.
        self.remove_list = remove_list
        self.full_image = False

        self.lost_found = '/lost+found'

        self.removed_file = TAR_FILE_REMOVED_FILE

        # Identify the directory the files the command file refers to
        # live in.
        if self.options.cmdfile:
            self.file_dir = os.path.dirname(self.options.cmdfile)

        # list of systemd unit D-Bus paths (such as
        # '/org/freedesktop/systemd1/unit/cron_2eservice') that need to
        # be restarted since they have running processes that are
        # holding open files that the upgrader needs to replace.
        self.services = []

        # If True, the other partition is considered empty. In this
        # scenario, prior to unpacking the latest image, the _current_
        # rootfs image is copied to the other partition.
        self.other_is_empty = False

        # True if the upgrader has reformatted the "other" partition.
        self.other_has_been_formatted = False

        # cache_dir is used as a scratch pad, for downloading new images
        # to and bind mounting the rootfs.
        if self.options.tmpdir:
            self.cache_dir = self.options.tmpdir
        else:
            self.cache_dir = self.DEFAULT_CACHE_DIR

    def update_timestamp(self):
        '''
        Update the timestamp file to record the time the last upgrade
        completed successfully.
        '''
        file = os.path.join(self.cache_dir, self.TIMESTAMP_FILE)
        touch_file(file)

    def get_mount_target(self):
        '''
        Get the full path to the mount target directory.
        '''
        return os.path.join(self.cache_dir, self.MOUNT_TARGET)

    def prepare(self):
        '''
        Required setup.
        '''

        target = self.get_mount_target()
        if subprocess.call([self.MOUNTPOINT_CMD, "-q", target]) != 0:
            raise Exception(
                "The {} directory is not a mountpoint".format(target))

        if self.options.dry_run:
            return

        if self.options.root_dir != '/':
            # Don't modify root when running in test mode
            return

    def run(self):
        '''
        Execute the commands in the command file
        '''
        self.prepare()

        for cmdline in self.commands:
            cmdline = cmdline.strip()

            args = cmdline.split()
            cmd = args[0]
            args = args[1:]

            if self.dispatcher[cmd]:
                log.debug('running dispatcher {} ({})'.format(cmd, args))
                self.dispatcher[cmd](args)
            else:
                log.warning('ignoring bogus input line: {}'
                            .format(cmdline))
        self.finish()

    def finish(self):
        '''
        Final tidy-up.
        '''

        if self.options.leave_files:
            log.debug('not removing files')
        else:
            for file in self.remove_list:
                log.debug('removing file {}'.format(file))
                os.remove(file)

        # Don't remount or reboot in test mode.
        if self.options.root_dir != '/':
            return

        self.update_timestamp()

    def _cmd_format(self, args):
        try:
            target = args[0]
        except:
            log.warning('expected target')
            return

        if target == 'system':
            self.full_image = True

        # Don't modify the system state
        if self.options.dry_run or self.options.root_dir != '/':
            return

        other = self.get_mount_target()

        source, fstype, label = get_mount_details(other)

        unmount(other)
        mkfs(source, fstype, label)

        # leave the mount as it was initially.
        mount(source, other, "ro")

        self.other_has_been_formatted = True

    def _cmd_load_keyring(self, args):
        try:
            keyring = args[0]
            signature = args[1]
        except:
            log.warning('expected keyring and signature')
            return

        keyring = os.path.join(self.file_dir, keyring)
        signature = os.path.join(self.file_dir, signature)

        self.remove_list.append(keyring)
        self.remove_list.append(signature)

        # already handled by system-image-cli(1) on image download
        log.info('ignoring keyring {} and signature {}'
                 ' (already handled)'
                 .format(keyring, signature))

    def _cmd_mount(self, args):
        """
        Although the name of this method matches the verb in the command
        file, it actually remounts "other" r/w now since dual partition
        systems have the other partition permanently mounted read-only.
        """

        try:
            target_type = args[0]
        except:
            log.warning('expected target type')
            return

        # the mount command is an imperative one: this script decides
        # what to mount and where :)
        if target_type != 'system':
            log.warning('unknown mount target type: {}'
                        .format(target_type))
            return

        if self.other_considered_empty():
            log.debug('other root partition is empty')
            self.other_is_empty = True

        if self.other_is_empty and not self.other_has_been_formatted:
            # We believe "other" is empty. However, it's possible that a
            # previous attempt at upgrading failed due to a power
            # outage. In that scenario, "other" may contain most of a
            # rootfs image, but just be missing the files used to
            # determine that the partition is empty. As such, if we
            # believe the partition is empty, it should forcibly be made
            # empty since it may contain detritus from a
            # previously-failed unpack (possibly caused by a power
            # outage).
            log.warning('reformatting other (no system image)')
            self._cmd_format("system")

        self.remount_rootfs(writable=True)

        if self.other_is_empty and not self.full_image:
            # Copy current rootfs data to the other rootfs's blank
            # partition.
            log.debug('syncing root partitions')
            self.sync_partitions()

    def _cmd_unmount(self, args):
        """
        Although the name of this method matches the verb in the command
        file, it actually remounts "other" r/o now since dual partition
        systems have the other partition permanently mounted read-only.
        """

        try:
            target_type = args[0]
        except:
            log.warning('expected target type')
            return

        # the unmount command is an imperative one: this script decides
        # what to mount and where :)
        if target_type != 'system':
            log.warning('unknown mount target type: {}'
                        .format(target_type))
            return

        self.remount_rootfs(writable=False)

    def other_considered_empty(self):
        '''
        Returns True if the other rootfs should be considered empty.

        See sync_partitions() for further details.

        '''
        target = self.get_mount_target()
        channel_ini = os.path.normpath('{}/{}'.format(target,
                                       SYSTEM_IMAGE_CHANNEL_CONFIG))

        try:
            st = os.stat(channel_ini)
        except OSError:
            # ENOENT
            return True

        if st.st_size == 0:
            # The config file is zero bytes long. This is used as a
            # marker to denote that the initial "sync system-a to
            # system-b" was started, but was interrupted, either in the
            # copy phase or in the s-i unpack phase. Either way, the
            # partition must be considered empty until the config file
            # has a size > 0.
            return True

        return False

    def remount_rootfs(self, writable=False):
        target = self.get_mount_target()

        if writable:
            root, _, _ = get_mount_details(target)

            # ro->rw so need to fsck first.
            unmount(target)

            # needs to be mounted writable, so check it first!
            fsck(root)

            mount(root, target, "rw")
        else:
            # rw->ro so no fsck required.
            remount(target, "ro")

    def get_file_contents(self, tar, file):
        '''
        @tar: tarfile object.
        @file: full path to file within @tar to extract.

        Returns: contents of @file from within @tar.
        '''
        tmpdir = tempfile.mkdtemp(prefix=script_name)
        tar.extract(path=tmpdir, member=tar.getmember(file))

        path = os.path.join(tmpdir, file)

        lines = []

        with open(path, 'r') as f:
            lines = f.readlines()

        lines = [line.rstrip() for line in lines]

        shutil.rmtree(tmpdir)

        return lines

    def _cmd_update(self, args):
        '''
        Unpack a new system image.

        Note that this method will be called multiple times, once for
        each file that comprises the upgrade.
        '''
        try:
            file = args[0]
            signature = args[1]
        except:
            log.warning('expected file and signature')
            return

        file = os.path.join(self.file_dir, file)
        signature = os.path.join(self.file_dir, signature)

        for f in (file, signature):
            if not os.path.exists(f):
                log.warning('ignoring missing file {}'
                            .format(f))
                return

            self.remove_list.append(f)

        log.info('applying update: {}'.format(file))
        tar = tarfile.open(file)

        found_removed_file = False

        # start with a list of all the files in the tarfile
        all_files = tar.getnames()

        if self.removed_file in all_files:
            # Exclude the removed file from the normal list as it is not
            # to be unpacked on the normal filesystem.
            all_files.remove(self.removed_file)
            found_removed_file = True

        # convert the paths back into real filesystem paths
        # (by dropping the 'system/' prefix).
        all_files = list(map(remove_prefix, all_files))

        # remove entries for files which failed the remove_prefix() check
        all_files = [x for x in all_files if x is not None]

        if found_removed_file:
            to_remove = self.get_file_contents(tar, self.removed_file)
        else:
            to_remove = []

        # by definition, full images don't have 'removed' files.
        if not self.full_image:
            if found_removed_file:
                log.debug('processing {} file'
                          .format(self.removed_file))

                # process backwards to work around bug LP:#1381134.
                for remove in sorted(to_remove, reverse=True):

                    if not remove:
                        # handle invalid removed file entry (see LP:#1437225)
                        continue

                    remove = remove.strip()

                    # don't remove devices
                    if remove.startswith('{}dev/'
                                         .format(TAR_FILE_SYSTEM_PREFIX)):
                        continue

                    # The upgrader runs as root so can modify any
                    # file. However, we should still check to ensure the
                    # paths look "reasonable". Since the server
                    # should only ever specify absolute paths,
                    # ignore anything that isn't.
                    if '../' in remove:
                        continue

                    final = os.path.join(self.cache_dir, remove)

                    if not os.path.exists(final):
                        # This scenario can only mean there is a bug
                        # with system-image generation (or someone made
                        # the image writable and removed some files
                        # manually).
                        log.debug('ignoring non-existent file {}'
                                  .format(final))
                        continue

                    if self.options.dry_run:
                        log.info('DRY-RUN: would remove file {}'.format(final))
                        continue

                    log.debug('removing file {}'.format(final))
                    try:
                        if os.path.isdir(final) and not os.path.islink(final):
                            shutil.rmtree(final)
                        else:
                            os.remove(final)
                    except Exception as e:
                        log.warning('failed to remove {}: {}'
                                    .format(final, e))

        if self.options.dry_run:
            log.info('DRY-RUN: would apply the following files:')
            tar.list(verbose=True)
        else:
            log.debug('starting unpack')
            # see bug #1408579, this forces tarfile to use tarinfo.gid
            # instead of looking up the gid by local name which might
            # be incorrect
            from unittest.mock import patch
            with patch("grp.getgrnam") as m:
                m.side_effect = KeyError()
                tar.extractall(path=self.cache_dir,
                               members=tar_generator(
                                   tar, self.cache_dir,
                                   self.removed_file,
                                   self.options.root_dir))
        tar.close()

        os.sync()

    def sync_partitions(self):
        '''
        Copy all rootfs data from the current partition to the other
        partition.

        XXX: Assumes that the other rootfs is already mounted r/w to
        mountpoint get_mount_target().

        Strategy: The partition sync is only called once (when the "other"
        rootfs is empty), to make the empty partition identical to the
        current rootfs partition, before unpacking the latest s-i updates
        onto it.

        Since there are 2 operations to perform for the initial update
        (copy and unpack) and since the upgrader must be resilient to
        power outages at any point when either of the operations are
        running, we need a way for the upgrader to detect if a previous
        invocation failed. This is achieved by treating the s-i
        configuration file (SYSTEM_IMAGE_CHANNEL_CONFIG) in a special
        way.

        = Background =

        A rootfs is not considered complete until the file
        SYSTEM_IMAGE_CHANNEL_CONFIG has been unpacked onto it. However,
        that implies we cannot copy the original
        SYSTEM_IMAGE_CHANNEL_CONFIG from the current rootfs to "other"
        as the copy is only the first operation and we don't want that
        "partial update" to be considered as valid.

        Ideally, we would perform the copy, excluding
        just SYSTEM_IMAGE_CHANNEL_CONFIG. However, rsync (which has
        excellent support for copying files with exclusions) is not
        available.

        = Approach =

        The current approach is:

            1) Create the directory part of SYSTEM_IMAGE_CHANNEL_CONFIG
               on "other".

            2) Touch SYSTEM_IMAGE_CHANNEL_CONFIG as an empty file on
               "other".

            3) Perform an "update copy" from current to "other". Note
               that this will not modify the pre-existing
               SYSTEM_IMAGE_CHANNEL_CONFIG, but will conveniently correct
               the permissions on /etc/system-image/ to match those on
               "current" (in the case that the original permissions used
               by this function do not match those on "current").

        The second part of the operation is to unpack the s-i update onto
        "other" (the final part of which is to unpack
        SYSTEM_IMAGE_CHANNEL_CONFIG (which will be >0 bytes), thus
        making that rootfs complete and valid).

        = Special treatment of SYSTEM_IMAGE_CHANNEL_CONFIG =

        Using this approach makes it easy to determine if the initial
        update failed since if either SYSTEM_IMAGE_CHANNEL_CONFIG does
        not exist or is only zero bytes, the s-i unpack must have failed
        meaning the rootfs must be considered "empty".
        '''
        target = self.get_mount_target()
        bindmount_rootfs_dir = tempfile.mkdtemp(prefix=script_name,
                                                dir=self.cache_dir)
        bind_mount("/", bindmount_rootfs_dir)

        cwd = os.getcwd()
        os.chdir(bindmount_rootfs_dir)

        channel_ini = os.path.normpath('{}/{}'.format(target,
                                       SYSTEM_IMAGE_CHANNEL_CONFIG))
        si_dir = os.path.dirname(channel_ini)
        os.makedirs(si_dir)
        touch_file(channel_ini)

        args = ['/bin/cp', '-au', '.', target]

        log.debug('running (from directory {}): {}'
                  .format(bindmount_rootfs_dir, args))

        proc = subprocess.Popen(args,
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)
        if proc.wait() != 0:
            os.chdir(cwd)
            log.error('failed to sync partitions')
            sys.exit(1)

        os.sync()

        os.chdir(cwd)

        unmount(bindmount_rootfs_dir)
        os.rmdir(bindmount_rootfs_dir)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# --------------------------------------------------------------------
# Copyright © 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

# --------------------------------------------------------------------
# Functional tests for the snappy upgrader.
# --------------------------------------------------------------------

import sys
import os
import logging
import unittest

from ubuntucoreupgrader.upgrader import (
    Upgrader,
    parse_args,
)

base_dir = os.path.abspath(os.path.dirname(__file__))
module_dir = os.path.normpath(os.path.realpath(base_dir + os.sep + '..'))
sys.path.append(base_dir)

from ubuntucoreupgrader.tests.utils import (
    append_file,
    TEST_DIR_MODE,
    create_file,
    UbuntuCoreUpgraderTestCase,
    )

CMD_FILE = 'ubuntu_command'


def call_upgrader(command_file, root_dir, update):
    '''
    Invoke the upgrader.

    :param command_file: commands file to drive the upgrader.
    :param root_dir: Test directory to apply the upgrade to.
    :param update: UpdateTree object.
    '''

    args = []
    args += ['--root-dir', root_dir]
    args += ['--debug', '1']

    # don't delete the archive and command files.
    # The tests clean up after themselves so they will get removed then,
    # but useful to have them around to diagnose test failures.
    args.append('--leave-files')

    args.append(command_file)
    commands = file_to_list(command_file)

    upgrader = Upgrader(parse_args(args), commands, [])
    upgrader.cache_dir = root_dir
    upgrader.MOUNTPOINT_CMD = "true"
    upgrader.run()


def create_device_file(path, type='c', major=-1, minor=-1):
    '''
    Create a device file.

    :param path: full path to device file.
    :param type: 'c' or 'b' (character or block).
    :param major: major number.
    :param minor: minor number.

    XXX: This doesn't actually create a device node,
    it simply creates a regular empty file whilst ensuring the filename
    gives the impression that it is a device file.

    This hackery is done for the following reasons:

    - non-priv users cannot create device nodes (and the tests run as a
      non-priv user).
    - the removed file in the upgrade tar file does not actually specify the
      _type_ of the files to remove. Hence, we can pretend that the file
      to remove is a device file since the upgrader cannot know for sure
      (it can check the existing on-disk file, but that isn't conclusive
      since the admin may have manually modified a file, or the server
      may have generated an invalid remove file - the upgrader cannot
      know for sure.
    '''
    assert (os.path.dirname(path).endswith('/dev'))

    append_file(path, 'fake-device file')


def create_directory(path):
    '''
    Create a directory.
    '''
    os.makedirs(path, mode=TEST_DIR_MODE, exist_ok=False)


def create_sym_link(source, dest):
    '''
    Create a symbolic link.

    :param source: existing file to link to.
    :param dest: name for the sym link.
    '''
    dirname = os.path.dirname(dest)
    os.makedirs(dirname, mode=TEST_DIR_MODE, exist_ok=True)

    os.symlink(source, dest)


def create_hard_link(source, dest):
    '''
    Create a hard link.

    :param source: existing file to link to.
    :param dest: name for the hard link.
    '''
    os.link(source, dest)


def is_sym_link_broken(path):
    '''
    :param path: symbolic link to check.
    :return: True if the specified path is a broken symbolic link,
     else False.
    '''
    try:
        os.lstat(path)
        os.stat(path)
    except:
        return True
    return False


def make_command_file(path, update_list):
    '''
    Create a command file that the upgrader processes.

    :param path: full path to file to create,
    :param update_list: list of update archives to include.
    '''
    l = []

    for file in update_list:
        l.append('update {} {}.asc'.format(file, file))

    # flatten
    contents = "\n".join(l) + '\n'

    append_file(path, contents)


def file_to_list(path):
    '''
    Convert the specified file into a list and return it.
    '''
    lines = []

    with open(path, 'r') as f:
        lines = f.readlines()

    lines = [line.rstrip() for line in lines]

    return lines


class UpgraderFileRemovalTestCase(UbuntuCoreUpgraderTestCase):
    '''
    Test how the upgrader handles the removals file.
    '''

    def test_remove_file(self):
        '''
        Ensure the upgrader can remove a regular file.
        '''

        file = 'a-regular-file'

        self.update.add_to_removed_file([file])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        file_path = os.path.join(self.victim_dir, file)
        create_file(file_path, 'foo bar')

        self.assertTrue(os.path.exists(file_path))
        self.assertTrue(os.path.isfile(file_path))

        # remove 'system' suffix that upgrader will add back on
        vdir = os.path.split(self.victim_dir)[0]

        call_upgrader(cmd_file, vdir, self.update)

        self.assertFalse(os.path.exists(file_path))

    def test_remove_directory(self):
        '''
        Ensure the upgrader can remove a directory.
        '''
        dir = 'a-directory'

        self.update.add_to_removed_file([dir])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        dir_path = os.path.join(self.victim_dir, dir)
        create_directory(dir_path)

        self.assertTrue(os.path.exists(dir_path))
        self.assertTrue(os.path.isdir(dir_path))

        vdir = os.path.split(self.victim_dir)[0]
        call_upgrader(cmd_file, vdir, self.update)

        self.assertFalse(os.path.exists(dir_path))

    def test_remove_sym_link_file(self):
        '''
        Ensure the upgrader can remove a symbolic link to a file.
        '''
        src = 'the-source-file'
        link = 'the-symlink-file'

        self.update.add_to_removed_file([link])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        src_file_path = os.path.join(self.victim_dir, src)
        link_file_path = os.path.join(self.victim_dir, link)

        create_file(src_file_path, 'foo bar')

        self.assertTrue(os.path.exists(src_file_path))
        self.assertTrue(os.path.isfile(src_file_path))
        self.assertFalse(os.path.islink(src_file_path))

        create_sym_link(src_file_path, link_file_path)
        self.assertTrue(os.path.exists(link_file_path))
        self.assertTrue(os.path.islink(link_file_path))

        vdir = os.path.split(self.victim_dir)[0]
        call_upgrader(cmd_file, vdir, self.update)

        # original file should still be there
        self.assertTrue(os.path.exists(src_file_path))
        self.assertTrue(os.path.isfile(src_file_path))
        self.assertFalse(os.path.islink(src_file_path))

        # link should have gone
        self.assertFalse(os.path.exists(link_file_path))

    def test_remove_sym_link_directory(self):
        '''
        Ensure the upgrader can remove a symbolic link to a directory.
        '''
        dir = 'the-source-directory'
        link = 'the-symlink-file'

        self.update.add_to_removed_file([link])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        src_dir_path = os.path.join(self.victim_dir, dir)
        link_file_path = os.path.join(self.victim_dir, link)

        create_directory(src_dir_path)

        self.assertTrue(os.path.exists(src_dir_path))
        self.assertTrue(os.path.isdir(src_dir_path))
        self.assertFalse(os.path.islink(src_dir_path))

        create_sym_link(src_dir_path, link_file_path)
        self.assertTrue(os.path.exists(link_file_path))
        self.assertTrue(os.path.islink(link_file_path))

        vdir = os.path.split(self.victim_dir)[0]
        call_upgrader(cmd_file, vdir, self.update)

        # original directory should still be there
        self.assertTrue(os.path.exists(src_dir_path))
        self.assertTrue(os.path.isdir(src_dir_path))
        self.assertFalse(os.path.islink(src_dir_path))

        # link should have gone
        self.assertFalse(os.path.exists(link_file_path))

    def test_remove_hardlink(self):
        '''
        Ensure the upgrader can remove a hard link to a file.
        '''
        src = 'the-source-file'
        link = 'the-hardlink-file'

        self.update.add_to_removed_file([link])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        src_file_path = os.path.join(self.victim_dir, src)
        link_file_path = os.path.join(self.victim_dir, link)

        create_file(src_file_path, 'foo bar')

        src_inode = os.stat(src_file_path).st_ino

        self.assertTrue(os.path.exists(src_file_path))
        self.assertTrue(os.path.isfile(src_file_path))

        create_hard_link(src_file_path, link_file_path)
        self.assertTrue(os.path.exists(link_file_path))

        link_inode = os.stat(link_file_path).st_ino

        self.assertTrue(src_inode == link_inode)

        vdir = os.path.split(self.victim_dir)[0]
        call_upgrader(cmd_file, vdir, self.update)

        # original file should still be there
        self.assertTrue(os.path.exists(src_file_path))
        self.assertTrue(os.path.isfile(src_file_path))

        # Inode should not have changed.
        self.assertTrue(os.stat(src_file_path).st_ino == src_inode)

        # link should have gone
        self.assertFalse(os.path.exists(link_file_path))

    def test_remove_device_file(self):
        '''
        Ensure the upgrader can deal with a device file.

        XXX: Note that The upgrader currently "deals" with them by
        XXX: ignoring them :-)

        '''
        file = '/dev/a-fake-device'

        self.update.add_to_removed_file([file])

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        file_path = '{}{}'.format(self.victim_dir, file)

        create_device_file(file_path)

        self.assertTrue(os.path.exists(file_path))

        # sigh - we can't assert the that filetype is a char/block
        # device because it won't be :)
        self.assertTrue(os.path.isfile(file_path))

        vdir = os.path.split(self.victim_dir)[0]
        call_upgrader(cmd_file, vdir, self.update)

        # upgrader doesn't currently remove device files
        self.assertTrue(os.path.exists(file_path))


class UpgraderFileAddTestCase(UbuntuCoreUpgraderTestCase):
    '''
    Test how the upgrader handles adding new files.
    '''

    def test_create_file(self):
        '''
        Ensure the upgrader can create a regular file.
        '''
        file = 'created-regular-file'

        file_path = os.path.join(self.update.system_dir, file)

        create_file(file_path, 'foo bar')

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        file_path = os.path.join(self.victim_dir, file)
        self.assertFalse(os.path.exists(file_path))

        call_upgrader(cmd_file, self.victim_dir, self.update)

        self.assertTrue(os.path.exists(file_path))
        self.assertTrue(os.path.isfile(file_path))

    def test_create_directory(self):
        '''
        Ensure the upgrader can create a directory.
        '''
        dir = 'created-directory'

        dir_path = os.path.join(self.update.system_dir, dir)

        create_directory(dir_path)

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        dir_path = os.path.join(self.victim_dir, dir)
        self.assertFalse(os.path.exists(dir_path))

        call_upgrader(cmd_file, self.victim_dir, self.update)

        self.assertTrue(os.path.exists(dir_path))
        self.assertTrue(os.path.isdir(dir_path))

    def test_create_absolute_sym_link_to_file(self):
        '''
        Ensure the upgrader can create a symbolic link to a file (which
        already exists and is not included in the update archive).
        '''
        src = 'the-source-file'
        link = 'the-symlink-file'

        # the file the link points to should *NOT* be below the 'system/'
        # directory (since there isn't one post-unpack).

        # an absolute sym-link target path
        src_file_path = '/{}'.format(src)
        link_file_path = os.path.join(self.update.system_dir, link)

        victim_src_file_path = os.path.normpath('{}/{}'
                                                .format(self.victim_dir,
                                                        src_file_path))
        victim_link_file_path = os.path.join(self.victim_dir, link)

        # Create a broken sym link ('/system/<link> -> /<src>')
        create_sym_link(src_file_path, link_file_path)

        self.assertTrue(os.path.lexists(link_file_path))
        self.assertTrue(os.path.islink(link_file_path))
        self.assertTrue(is_sym_link_broken(link_file_path))

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        create_file(victim_src_file_path, 'foo')

        self.assertTrue(os.path.exists(victim_src_file_path))
        self.assertTrue(os.path.isfile(victim_src_file_path))

        self.assertFalse(os.path.exists(victim_link_file_path))

        call_upgrader(cmd_file, self.victim_dir, self.update)

        self.assertTrue(os.path.exists(victim_src_file_path))
        self.assertTrue(os.path.isfile(victim_src_file_path))

        # upgrader should have created the link in the victim directory
        self.assertTrue(os.path.lexists(victim_link_file_path))
        self.assertTrue(os.path.islink(victim_link_file_path))
        self.assertFalse(is_sym_link_broken(victim_link_file_path))

    def test_create_relative_sym_link_to_file(self):
        '''
        Ensure the upgrader can create a symbolic link to a file (which
        already exists and is not included in the update archive).
        '''
        src = 'a/b/c/the-source-file'
        link = 'a/d/e/the-symlink-file'

        # a relative sym-link target path
        # ##src_file_path = '../../b/c/{}'.format(src)
        src_file_path = '../../b/c/the-source-file'.format(src)

        # the file the link points to should *NOT* be below the 'system/'
        # directory (since there isn't one post-unpack).

        link_file_path = os.path.join(self.update.system_dir, link)

        victim_src_file_path = os.path.normpath('{}/{}'
                                                .format(self.victim_dir,
                                                        src))
        victim_link_file_path = os.path.join(self.victim_dir, link)

        create_sym_link(src_file_path, link_file_path)

        self.assertTrue(os.path.lexists(link_file_path))
        self.assertTrue(os.path.islink(link_file_path))
        self.assertTrue(is_sym_link_broken(link_file_path))

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        create_file(victim_src_file_path, 'foo')

        self.assertTrue(os.path.exists(victim_src_file_path))
        self.assertTrue(os.path.isfile(victim_src_file_path))

        self.assertFalse(os.path.exists(victim_link_file_path))

        call_upgrader(cmd_file, self.victim_dir, self.update)

        self.assertTrue(os.path.exists(victim_src_file_path))
        self.assertTrue(os.path.isfile(victim_src_file_path))

        # upgrader should have created the link in the victim directory
        self.assertTrue(os.path.lexists(victim_link_file_path))
        self.assertTrue(os.path.islink(victim_link_file_path))
        self.assertFalse(is_sym_link_broken(victim_link_file_path))

    def test_create_broken_sym_link_file(self):
        '''
        Ensure the upgrader can create a broken symbolic link
        (one that points to a non-existent file).
        '''
        src = 'the-source-file'
        link = 'the-symlink-file'

        # the file the link points to should *NOT* be below the 'system/'
        # directory (since there isn't one post-unpack).
        src_file_path = src

        link_file_path = os.path.join(self.update.system_dir, link)

        # Create a broken sym link ('/system/<link> -> /<src>')
        create_sym_link(src_file_path, link_file_path)

        self.assertTrue(os.path.lexists(link_file_path))
        self.assertTrue(os.path.islink(link_file_path))
        self.assertTrue(is_sym_link_broken(link_file_path))

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        victim_src_file_path = os.path.join(self.victim_dir, src)
        victim_link_file_path = os.path.join(self.victim_dir, link)

        self.assertFalse(os.path.exists(victim_src_file_path))
        self.assertFalse(os.path.exists(victim_link_file_path))

        call_upgrader(cmd_file, self.victim_dir, self.update)

        # source still shouldn't exist
        self.assertFalse(os.path.exists(victim_src_file_path))

        # upgrader should have created the link in the victim directory
        self.assertTrue(os.path.lexists(victim_link_file_path))
        self.assertTrue(os.path.islink(victim_link_file_path))
        self.assertTrue(is_sym_link_broken(victim_link_file_path))


class UpgraderRemoveFileTests(UbuntuCoreUpgraderTestCase):

    def common_removed_file_test(self, contents):
        '''
        Common code to test for an invalid removed file.

        The contents parameter specifies the contents of the removed
        file.
        '''
        file = 'created-regular-file'

        create_file(self.update.removed_file, contents)

        file_path = os.path.join(self.update.system_dir, file)
        create_file(file_path, 'foo bar')

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        cmd_file = os.path.join(self.update.tmp_dir, CMD_FILE)
        make_command_file(cmd_file, [self.TARFILE])

        vdir = os.path.split(self.victim_dir)[0]

        file_path = os.path.join(vdir, file)
        self.assertFalse(os.path.exists(file_path))

        # XXX: There is an implicit test here since if the upgrader
        # fails (as documented on LP: #1437225), this test will also
        # fail.
        call_upgrader(cmd_file, vdir, self.update)

        self.assertTrue(os.path.exists(vdir))
        self.assertTrue(os.path.exists(self.victim_dir))
        self.assertTrue(os.path.exists(file_path))

        # ensure the empty removed file hasn't removed the directory the
        # unpack applies to.
        self.assertTrue(self.victim_dir)

    def test_removed_file_empty(self):
        '''
        Ensure the upgrader ignores an empty 'removed' file.
        '''
        self.common_removed_file_test('')

    def test_removed_file_space(self):
        '''
        Ensure the upgrader handles a 'removed' file containing just a
        space.
        '''
        self.common_removed_file_test(' ')

    def test_removed_file_nl(self):
        '''
        Ensure the upgrader handles a 'removed' file containing just a
        newline
        '''
        self.common_removed_file_test('\n')


def main():
    kwargs = {}
    format =             \
        '%(asctime)s:'   \
        '%(filename)s:'  \
        '%(name)s:'      \
        '%(funcName)s:'  \
        '%(levelname)s:' \
        '%(message)s'

    kwargs['format'] = format

    # We want to see what's happening
    kwargs['level'] = logging.DEBUG

    logging.basicConfig(**kwargs)

    unittest.main(
        testRunner=unittest.TextTestRunner(
            stream=sys.stdout,
            verbosity=2,

            # don't keep running tests if one fails
            # (... who _wouldn't_ want this???)
            failfast=True
        ),

    )

if __name__ == '__main__':
    main()
